package ru.dbotthepony.mc.otm.block.entity.matter

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.block.entity.MatteryWorkerBlockEntity
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.IMatterHandler
import ru.dbotthepony.mc.otm.capability.matter.MatterDirection
import ru.dbotthepony.mc.otm.capability.matter.MatterHandlerImpl
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.MatteryContainerFilter
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.IMatterGraphNode
import ru.dbotthepony.mc.otm.graph.matter.MatterNetworkGraph
import ru.dbotthepony.mc.otm.item.MatterDustItem
import ru.dbotthepony.mc.otm.core.map
import ru.dbotthepony.mc.otm.menu.MatterRecyclerMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.container.set
import ru.dbotthepony.mc.otm.core.set

class MatterRecyclerBlockEntity(blockPos: BlockPos, blockState: BlockState)
	: MatteryWorkerBlockEntity<MatteryWorkerBlockEntity.Job>(MBlockEntities.MATTER_RECYCLER, blockPos, blockState, ::Job), IMatterGraphNode {

	val matter = MatterHandlerImpl(
		this::matterLevelUpdated,
		MatterDirection.EXTRACT,
		STORAGE
	)

	val container = MatteryContainer(this::itemContainerUpdated, 1)
	override val matterNode = Graph6Node<IMatterGraphNode>(this)
	private var resolverNode = LazyOptional.of { this }
	private var valid = true
	override val energy = WorkerEnergyStorage(this::powerLevelUpdated, MAX_POWER)

	override fun getMatterHandler(): IMatterHandler {
		return matter
	}

	private val itemHandler = container.handler(object : MatteryContainerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return stack.item is MatterDustItem
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return false
		}
	})

	override fun invalidateCaps() {
		super.invalidateCaps()
		matter.invalidate()
		itemHandler.invalidate()
		resolverNode.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		matter.revive()
		itemHandler.revive()
		resolverNode = LazyOptional.of { this }
	}

	override fun setRemoved() {
		super.setRemoved()
		matterNode.destroy(::MatterNetworkGraph)
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		if (p_155231_ is ServerLevel)
			MatterNetworkGraph.discoverFull(this, matterNode)
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["matter"] = matter.serializeNBT()
		nbt["container"] = container.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		nbt.map("matter", matter::deserializeNBT)
		container.deserializeNBT(nbt["container"])
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (!valid)
			return super.getCapability(cap, side)

		return when (cap) {
			ForgeCapabilities.ITEM_HANDLER -> itemHandler.get().cast()
			MatteryCapability.MATTER -> matter.get().cast()
			MatteryCapability.MATTER_NODE -> resolverNode.cast()
			else -> super.getCapability(cap, side)
		}
	}

	override val defaultDisplayName: Component
		get() = MACHINE_NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterRecyclerMenu(containerID, inventory, this)
	}

	override fun onJobFinish(job: Job): Status {
		// вся логика в onWorkTick
		return Status.SUCCESS
	}

	override fun computeNextJob(): Pair<Job?, IdleReason?> {
		if (matter.missingMatter.isZero)
			return null to IdleReason.ITEM

		val stack = container[0]

		if (stack.isEmpty || stack.item !is MatterDustItem) {
			return null to IdleReason.ITEM
		}

		val dustMatter = (stack.item as MatterDustItem).getMatterValue(stack.copy().also { it.count = 1 }) ?: return null to IdleReason.ITEM
		stack.shrink(1)
		container.setChanged(0)
		return Job(dustMatter.value.toDouble() * MATTER_TICKS, POWER_CONSUMPTION) to null
	}

	override fun onWorkTick(requiredPower: ImpreciseFraction, extractedPower: ImpreciseFraction, ticksAdvanced: Double): Status {
		if ((level?.random?.nextDouble() ?: 1.0) <= 0.4)
			return Status.SUCCESS

		val receive = MATTER_PER_TICK * ticksAdvanced
		val received = matter.receiveMatterInner(receive, true)

		if (receive != received)
			return Status.FAILURE_MATTER

		matter.receiveMatterInner(receive, false)
		return Status.SUCCESS
	}

	fun tick() {
		basicTicker()

		val graph = matterNode.graph as MatterNetworkGraph? ?: return
		val received = graph.receiveMatter(matter.storedMatter, false)

		if (!received.isZero) {
			matter.extractMatterInner(received, false)
		}
	}

	companion object {
		private val MACHINE_NAME = TranslatableComponent("block.overdrive_that_matters.matter_recycler")
		private val STORAGE = ImpreciseFraction(40)
		private val POWER_CONSUMPTION = ImpreciseFraction(80)
		private val MAX_POWER = ImpreciseFraction(80_000)
		private const val MATTER_TICKS = 200.0
		private val MATTER_PER_TICK = ImpreciseFraction(1.0 / MATTER_TICKS)
	}
}
