package ru.dbotthepony.mc.otm.client.render

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.resources.ResourceLocation
import ru.dbotthepony.mc.otm.client.screen.panels.DockProperty

@Suppress("unused")
class SkinGrid(
	val texture: ResourceLocation,
	val width: Float,
	val height: Float,
	val rows: Int = 16,
	val columns: Int = 16,
) {
	var row = 0
	var column = 0

	val imageWidth get() = width * columns
	val imageHeight get() = height * rows

	fun skip(): SkinGrid {
		column++

		if (column >= columns) {
			jump()
		}

		return this
	}

	fun jump(): SkinGrid {
		row++
		column = 0
		return this
	}

	fun next(): SkinElement {
		val element = SkinElement(texture, column * width, row * height, width, height, imageWidth, imageHeight)
		skip()
		return element
	}

	fun reset(): SkinGrid {
		row = 0
		column = 0
		return this
	}

	fun seek(row: Int, column: Int): SkinGrid {
		this.row = row
		this.column = column
		return this
	}

	operator fun get(column: Int, row: Int) = SkinElement(texture, column * width, row * height, width, height, imageWidth, imageHeight)
}

fun ResourceLocation.element(
	x: Float,
	y: Float,
	w: Float,
	h: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, w, h, textureWidth, textureHeight)

fun ResourceLocation.pixel(
	x: Float,
	y: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, 1f, 1f, textureWidth, textureHeight)

fun ResourceLocation.hLine(
	x: Float,
	y: Float,
	width: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, width, 1f, textureWidth, textureHeight)

fun ResourceLocation.vLine(
	x: Float,
	y: Float,
	height: Float,
	textureWidth: Float = 256f,
	textureHeight: Float = 256f
) = SkinElement(this, x, y, 1f, height, textureWidth, textureHeight)

@Suppress("unused")
class SkinElement @JvmOverloads constructor(
	val texture: ResourceLocation,
	val x: Float,
	val y: Float,
	val w: Float,
	val h: Float,
	val imageWidth: Float = 256f,
	val imageHeight: Float = 256f
) {
	init {
		require(x >= 0f) { "Invalid x $x" }
		require(y >= 0f) { "Invalid y $y" }
		require(w > 0f) { "Invalid width $w" }
		require(h > 0f) { "Invalid height $h" }
		require(imageWidth > 0f) { "Invalid image width $imageWidth" }
		require(imageHeight > 0f) { "Invalid image height $imageHeight" }
	}

	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = w,
		height: Float = h,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRaw(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun render(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = w.toDouble(),
		height: Double = h.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) = render(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = w,
		height: Float = h,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) {
		RenderSystem.setShaderTexture(0, texture)
		RenderSystem.enableBlend()
		RenderSystem.defaultBlendFunc()

		renderRawPartial(stack, x, y, width, height, winding)
	}

	@JvmOverloads
	fun renderPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = w.toDouble(),
		height: Double = h.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) = renderPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderW(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = w,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1,
	) {
		render(stack, x, y, width = width, winding = winding)
	}

	@JvmOverloads
	fun renderW(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = w.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1,
	) = renderW(stack, x.toFloat(), y.toFloat(), width.toFloat(), winding)

	@JvmOverloads
	fun renderH(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		height: Float = h,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1,
	) {
		render(stack, x, y, height = height, winding = winding)
	}

	@JvmOverloads
	fun renderH(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		height: Double = h.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1,
	) = renderH(stack, x.toFloat(), y.toFloat(), height.toFloat(), winding)

	private val u0 = this.x / imageWidth
	private val v0 = this.y / imageHeight
	private val u1 = (this.x + w) / imageWidth
	private val v1 = (this.y + h) / imageHeight

	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = w,
		height: Float = h,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) {
		val winded = winding.translate(u0, v0, u1, v1)

		drawTexturedRect(
			stack,
			x,
			y,
			width,
			height,
			winded,
		)
	}

	@JvmOverloads
	fun renderRaw(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = w.toDouble(),
		height: Double = h.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) = renderRaw(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float = w,
		height: Float = h,
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) {
		val u1 = (this.x + width) / imageWidth
		val v1 = (this.y + height) / imageHeight

		val winded = winding.translate(u0, v0, u1, v1)

		drawTexturedRect(
			stack,
			x,
			y,
			width,
			height,
			winded,
		)
	}

	@JvmOverloads
	fun renderRawPartial(
		stack: PoseStack,
		x: Double,
		y: Double = 0.0,
		width: Double = w.toDouble(),
		height: Double = h.toDouble(),
		winding: UVWindingOrder = UVWindingOrder.U0_V0_U1_V1
	) = renderRawPartial(stack, x.toFloat(), y.toFloat(), width.toFloat(), height.toFloat(), winding)
}

class StretchingRectangleElement(
	val topLeft: SkinElement,
	val topRight: SkinElement,
	val bottomLeft: SkinElement,
	val bottomRight: SkinElement,
	val left: SkinElement,
	val right: SkinElement,
	val top: SkinElement,
	val bottom: SkinElement,
	val middle: SkinElement,
	val padding: DockProperty = DockProperty.EMPTY
) {
	fun render(
		stack: PoseStack,
		x: Float = 0f,
		y: Float = 0f,
		width: Float,
		height: Float,
	) {
		val bgWidth = width - topLeft.w.coerceAtMost(bottomLeft.w) - topRight.w.coerceAtMost(bottomRight.w) - padding.left - padding.right
		val bgHeight = height - topLeft.h.coerceAtMost(bottomLeft.h) - topRight.h.coerceAtMost(bottomRight.h) - padding.top - padding.bottom

		if (bgWidth > 0 && bgHeight > 0) {
			middle.render(stack, x + topLeft.w + padding.left, y + topLeft.h + padding.top, bgWidth, bgHeight)
		}

		top.render(stack, x + topLeft.w, y, width = width - topLeft.w - topRight.w)
		bottom.render(stack, x + bottomLeft.w, y + height - bottom.h, width = width - bottomLeft.w - bottomRight.w)

		left.render(stack, x, y + topLeft.h, height = height - topLeft.h - bottomLeft.h)
		right.render(stack, x + width - right.w, y + topRight.h, height = height - topRight.h - bottomRight.h)

		topLeft.render(stack, x, y)
		topRight.render(stack, x + width - topRight.w, y)

		bottomLeft.render(stack, x, y + height - bottomLeft.h)
		bottomRight.render(stack, x + width - bottomRight.w, y + height - bottomLeft.h)
	}
}

