package ru.dbotthepony.mc.otm.registry.objects

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import net.minecraft.world.item.BlockItem
import net.minecraft.world.item.DyeColor
import net.minecraft.world.item.Item
import net.minecraft.world.level.block.Block
import net.minecraftforge.registries.DeferredRegister
import net.minecraftforge.registries.RegistryObject
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.registry.LazyList
import ru.dbotthepony.mc.otm.registry.LazyMap
import java.util.EnumMap

@Suppress("PropertyName", "unused", "ReplaceGetOrSet", "ReplacePutWithAssignment")
class StripedColoredDecorativeBlock(
	val basename: String,
	val blockFactory: (colorA: DyeColor, colorB: DyeColor) -> Block,
	val itemFactory: ((colorA: DyeColor, colorB: DyeColor, block: Block) -> Item) = { _, _, block ->
		BlockItem(block, Item.Properties().tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB_DECORATIVE).stacksTo(64))
	}
) {
	private val mapBlocks = EnumMap<DyeColor, EnumMap<DyeColor, RegistryObject<Block>>>(DyeColor::class.java)
	private val mapItems = EnumMap<DyeColor, EnumMap<DyeColor, RegistryObject<Item>>>(DyeColor::class.java)

	fun getBlockNullable(base: DyeColor, stripe: DyeColor): Block? {
		if (!registeredBlocks) {
			throw IllegalStateException("Not yet registered blocks")
		}

		return mapBlocks.get(base)?.get(stripe)?.get()
	}

	fun getItemNullable(base: DyeColor, stripe: DyeColor): Item? {
		if (!registeredItems) {
			throw IllegalStateException("Not yet registered items")
		}

		return mapItems.get(base)?.get(stripe)?.get()
	}

	fun getBlock(base: DyeColor, stripe: DyeColor): Block {
		if (!registeredBlocks) {
			throw IllegalStateException("Not yet registered blocks")
		}

		return mapBlocks.get(base)?.get(stripe)?.get() ?: throw NoSuchElementException("$base colored $stripe striped block does not exist")
	}

	fun getItem(base: DyeColor, stripe: DyeColor): Item {
		if (!registeredItems) {
			throw IllegalStateException("Not yet registered items")
		}

		return mapItems.get(base)?.get(stripe)?.get() ?: throw NoSuchElementException("$base colored $stripe striped item does not exist")
	}

	var registeredItems = false
		private set

	var registeredBlocks = false
		private set

	val items: List<Item> by lazy {
		if (!registeredItems) {
			throw IllegalStateException("Not yet registered items")
		}

		val build = ImmutableList.builder<() -> Item>()

		for (children in mapItems.values) {
			for (registryObject in children.values) {
				build.add(registryObject::get)
			}
		}

		LazyList(build.build())
	}

	val blocks: List<Block> by lazy {
		if (!registeredBlocks) {
			throw IllegalStateException("Not yet registered blocks")
		}

		val build = ImmutableList.builder<() -> Block>()

		for (children in mapBlocks.values) {
			for (registryObject in children.values) {
				build.add(registryObject::get)
			}
		}

		LazyList(build.build())
	}

	val mappedBlocks: Map<DyeColor, Map<DyeColor, Block>> by lazy {
		if (!registeredBlocks) {
			throw IllegalStateException("Not yet registered blocks")
		}

		val builder = ImmutableMap.Builder<DyeColor, Map<DyeColor, Block>>()

		for ((base, children) in mapBlocks) {
			val result = ArrayList<Pair<DyeColor, () -> Block>>()

			for ((stripe, registryObject) in children) {
				result.add(stripe to registryObject::get)
			}

			builder.put(base, LazyMap(result))
		}

		builder.build()
	}

	val mappedItems: Map<DyeColor, Map<DyeColor, Item>> by lazy {
		if (!registeredItems) {
			throw IllegalStateException("Not yet registered items")
		}

		val builder = ImmutableMap.Builder<DyeColor, Map<DyeColor, Item>>()

		for ((base, children) in mapItems) {
			val result = ArrayList<Pair<DyeColor, () -> Item>>()

			for ((stripe, registryObject) in children) {
				result.add(stripe to registryObject::get)
			}

			builder.put(base, LazyMap(result))
		}

		builder.build()
	}

	fun registerItems(registry: DeferredRegister<Item>) {
		for (base in DyeColor.values()) {
			for (stripe in DyeColor.values()) {
				if (base == stripe) {
					continue
				}

				mapItems.computeIfAbsent(base) { EnumMap(DyeColor::class.java) }
					.put(stripe, registry.register("${basename}_${base.name.lowercase()}_${stripe.name.lowercase()}") { itemFactory.invoke(base, stripe, mapBlocks.get(base)!!.get(stripe)!!.get()) })
			}
		}

		registeredItems = true
	}

	fun registerBlocks(registry: DeferredRegister<Block>) {
		for (base in DyeColor.values()) {
			for (stripe in DyeColor.values()) {
				if (base == stripe) {
					continue
				}

				mapBlocks.computeIfAbsent(base) { EnumMap(DyeColor::class.java) }
					.put(stripe, registry.register("${basename}_${base.name.lowercase()}_${stripe.name.lowercase()}") { blockFactory.invoke(base, stripe) })
			}
		}

		registeredBlocks = true
	}

	val blocksWithColor: List<Pair<Block, Pair<DyeColor, DyeColor>>> by lazy {
		if (!registeredBlocks) {
			throw IllegalStateException("Not yet registered blocks")
		}

		val build = ImmutableList.Builder<() -> Pair<Block, Pair<DyeColor, DyeColor>>>()

		for ((base, children) in mapBlocks) {
			for ((stripe, registryObject) in children) {
				build.add { registryObject.get() to (base to stripe) }
			}
		}

		LazyList(build.build())
	}

	val itemsWithColor: List<Pair<Item, Pair<DyeColor, DyeColor>>> by lazy {
		if (!registeredItems) {
			throw IllegalStateException("Not yet registered items")
		}

		val build = ImmutableList.Builder<() -> Pair<Item, Pair<DyeColor, DyeColor>>>()

		for ((base, children) in mapItems) {
			for ((stripe, registryObject) in children) {
				build.add { registryObject.get() to (base to stripe) }
			}
		}

		LazyList(build.build())
	}
}
